@@ -47,7 +47,8 @@ class ZookeeperJob(base.Job):
 
     def __init__(self, board, name, client, path,
                  uuid=None, details=None, book=None, book_data=None,
-                 created_on=None, backend=None):
+                 created_on=None, backend=None,
+                 priority=base.JobPriority.NORMAL):
         super(ZookeeperJob, self).__init__(board, name,
                                            uuid=uuid, details=details,
                                            backend=backend,
@@ -60,12 +61,17 @@ class ZookeeperJob(base.Job):
         basename = k_paths.basename(self._path)
         self._root = self._path[0:-len(basename)]
         self._sequence = int(basename[len(board.JOB_PREFIX):])
+        self._priority = priority
 
     @property
     def lock_path(self):
         """Path the job lock/claim and owner znode is stored."""
         return self._lock_path
 
+    @property
+    def priority(self):
+        return self._priority
+
     @property
     def path(self):
         """Path the job data znode is stored."""
@@ -175,15 +181,23 @@ class ZookeeperJob(base.Job):
         if not isinstance(other, ZookeeperJob):
             return NotImplemented
         if self.root == other.root:
-            return self.sequence < other.sequence
+            if self.priority == other.priority:
+                return self.sequence < other.sequence
+            else:
+                ordered = base.JobPriority.reorder(
+                    (self.priority, self), (other.priority, other))
+                if ordered[0] is self:
+                    return False
+                return True
         else:
+            # Different jobboards with different roots...
             return self.root < other.root
 
     def __eq__(self, other):
         if not isinstance(other, ZookeeperJob):
             return NotImplemented
-        return ((self.root, self.sequence) ==
-                (other.root, other.sequence))
+        return ((self.root, self.sequence, self.priority) ==
+                (other.root, other.sequence, other.priority))
 
     def __hash__(self):
         return hash(self.path)
@@ -285,16 +299,18 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
         self._emit_notifications = bool(emit_notifications)
         self._connected = False
 
-    def _emit(self, state, details):
+    def _try_emit(self, state, details):
         # Submit the work to the executor to avoid blocking the kazoo threads
         # and queue(s)...
         worker = self._worker
-        if worker is None:
+        if worker is None or not self._emit_notifications:
+            # Worker has been destroyed or we aren't supposed to emit anything
+            # in the first place...
             return
         try:
             worker.submit(self.notifier.notify, state, details)
         except RuntimeError:
-            # Notification thread is shutdown just skip submitting a
+            # Notification thread is/was shutdown just skip submitting a
             # notification...
             pass
 
@@ -321,11 +337,12 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
         if ensure_fresh:
             self._force_refresh()
         with self._job_cond:
-            return sorted(six.itervalues(self._known_jobs))
+            return sorted(six.itervalues(self._known_jobs), reverse=True)
 
     def _force_refresh(self):
         try:
-            children = self._client.get_children(self.path)
+            maybe_children = self._client.get_children(self.path)
+            self._on_job_posting(maybe_children, delayed=False)
         except self._client.handler.timeout_exception:
             excp.raise_with_cause(excp.JobFailure,
                                   "Refreshing failure, operation timed out")
@@ -337,14 +354,13 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
         except k_exceptions.KazooException:
             excp.raise_with_cause(excp.JobFailure,
                                   "Refreshing failure, internal error")
-        else:
-            self._on_job_posting(children, delayed=False)
 
     def iterjobs(self, only_unclaimed=False, ensure_fresh=False):
+        board_removal_func = lambda job: self._remove_job(job.path)
         return base.JobBoardIterator(
             self, LOG, only_unclaimed=only_unclaimed,
             ensure_fresh=ensure_fresh, board_fetch_func=self._fetch_jobs,
-            board_removal_func=lambda a_job: self._remove_job(a_job.path))
+            board_removal_func=board_removal_func)
 
     def _remove_job(self, path):
         if path not in self._known_jobs:
@@ -353,47 +369,59 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
             job = self._known_jobs.pop(path, None)
         if job is not None:
             LOG.debug("Removed job that was at path '%s'", path)
-            self._emit(base.REMOVAL, details={'job': job})
+            self._try_emit(base.REMOVAL, details={'job': job})
 
-    def _process_child(self, path, request):
+    def _process_child(self, path, request, quiet=True):
         """Receives the result of a child data fetch request."""
         job = None
         try:
             raw_data, node_stat = request.get()
             job_data = misc.decode_json(raw_data)
-            created_on = misc.millis_to_datetime(node_stat.ctime)
+            job_created_on = misc.millis_to_datetime(node_stat.ctime)
+            try:
+                job_priority = job_data['priority']
+                job_priority = base.JobPriority.convert(job_priority)
+            except KeyError:
+                job_priority = base.JobPriority.NORMAL
+            job_uuid = job_data['uuid']
+            job_name = job_data['name']
         except (ValueError, TypeError, KeyError):
-            LOG.warn("Incorrectly formatted job data found at path: %s",
-                     path, exc_info=True)
+            with excutils.save_and_reraise_exception(reraise=not quiet):
+                LOG.warn("Incorrectly formatted job data found at path: %s",
+                         path, exc_info=True)
         except self._client.handler.timeout_exception:
-            LOG.warn("Operation timed out fetching job data from path: %s",
-                     path, exc_info=True)
+            with excutils.save_and_reraise_exception(reraise=not quiet):
+                LOG.warn("Operation timed out fetching job data from path: %s",
+                         path, exc_info=True)
         except k_exceptions.SessionExpiredError:
-            LOG.warn("Session expired fetching job data from path: %s", path,
-                     exc_info=True)
+            with excutils.save_and_reraise_exception(reraise=not quiet):
+                LOG.warn("Session expired fetching job data from path: %s",
+                         path, exc_info=True)
         except k_exceptions.NoNodeError:
             LOG.debug("No job node found at path: %s, it must have"
                       " disappeared or was removed", path)
         except k_exceptions.KazooException:
-            LOG.warn("Internal error fetching job data from path: %s",
-                     path, exc_info=True)
+            with excutils.save_and_reraise_exception(reraise=not quiet):
+                LOG.warn("Internal error fetching job data from path: %s",
+                         path, exc_info=True)
         else:
             with self._job_cond:
                 # Now we can offically check if someone already placed this
                 # jobs information into the known job set (if it's already
                 # existing then just leave it alone).
                 if path not in self._known_jobs:
-                    job = ZookeeperJob(self, job_data['name'],
+                    job = ZookeeperJob(self, job_name,
                                        self._client, path,
                                        backend=self._persistence,
-                                       uuid=job_data['uuid'],
+                                       uuid=job_uuid,
                                        book_data=job_data.get("book"),
                                        details=job_data.get("details", {}),
-                                       created_on=created_on)
+                                       created_on=job_created_on,
+                                       priority=job_priority)
                     self._known_jobs[path] = job
                     self._job_cond.notify_all()
         if job is not None:
-            self._emit(base.POSTED, details={'job': job})
+            self._try_emit(base.POSTED, details={'job': job})
 
     def _on_job_posting(self, children, delayed=True):
         LOG.debug("Got children %s under path %s", children, self.path)
@@ -404,7 +432,6 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
                 # Skip lock paths or non-job-paths (these are not valid jobs)
                 continue
             child_paths.append(k_paths.join(self.path, c))
-
         # Figure out what we really should be investigating and what we
         # shouldn't (remove jobs that exist in our local version, but don't
         # exist in the children anymore) and accumulate all paths that we
@@ -441,15 +468,18 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
             if delayed:
                 request.rawlink(functools.partial(self._process_child, path))
             else:
-                self._process_child(path, request)
+                self._process_child(path, request, quiet=False)
 
-    def post(self, name, book=None, details=None):
+    def post(self, name, book=None, details=None,
+             priority=base.JobPriority.NORMAL):
         # NOTE(harlowja): Jobs are not ephemeral, they will persist until they
         # are consumed (this may change later, but seems safer to do this until
         # further notice).
+        job_priority = base.JobPriority.convert(priority)
         job_uuid = uuidutils.generate_uuid()
         job_posting = base.format_posting(job_uuid, name,
-                                          book=book, details=details)
+                                          book=book, details=details,
+                                          priority=job_priority)
         raw_job_posting = misc.binary_encode(jsonutils.dumps(job_posting))
         with self._wrap(job_uuid, None,
                         fail_msg_tpl="Posting failure: %s",
@@ -461,11 +491,12 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
             job = ZookeeperJob(self, name, self._client, job_path,
                                backend=self._persistence,
                                book=book, details=details, uuid=job_uuid,
-                               book_data=job_posting.get('book'))
+                               book_data=job_posting.get('book'),
+                               priority=job_priority)
             with self._job_cond:
                 self._known_jobs[job_path] = job
                 self._job_cond.notify_all()
-            self._emit(base.POSTED, details={'job': job})
+            self._try_emit(base.POSTED, details={'job': job})
             return job
 
     @base.check_who
@@ -569,14 +600,29 @@ class ZookeeperJobBoard(base.NotifyingJobBoard):
         entity_type = entity.kind
         if entity_type == c_base.Conductor.ENTITY_KIND:
             entity_path = k_paths.join(self.entity_path, entity_type)
-            self._client.ensure_path(entity_path)
-
-            conductor_name = entity.name
-            self._client.create(k_paths.join(entity_path,
-                                             conductor_name),
-                                value=misc.binary_encode(
-                                    jsonutils.dumps(entity.to_dict())),
-                                ephemeral=True)
+            try:
+                self._client.ensure_path(entity_path)
+                self._client.create(k_paths.join(entity_path, entity.name),
+                                    value=misc.binary_encode(
+                                        jsonutils.dumps(entity.to_dict())),
+                                    ephemeral=True)
+            except k_exceptions.NodeExistsError:
+                pass
+            except self._client.handler.timeout_exception:
+                excp.raise_with_cause(
+                    excp.JobFailure,
+                    "Can not register entity %s under %s, operation"
+                    " timed out" % (entity.name, entity_path))
+            except k_exceptions.SessionExpiredError:
+                excp.raise_with_cause(
+                    excp.JobFailure,
+                    "Can not register entity %s under %s, session"
+                    " expired" % (entity.name, entity_path))
+            except k_exceptions.KazooException:
+                excp.raise_with_cause(
+                    excp.JobFailure,
+                    "Can not register entity %s under %s, internal"
+                    " error" % (entity.name, entity_path))
         else:
             raise excp.NotImplementedError(
                 "Not implemented for other entity type '%s'" % entity_type)
