@@ -301,6 +301,36 @@ class EngineLinearFlowTest(utils.EngineTestBase):
         results = engine.storage.fetch_all()
         self.assertEqual(2, results['result'])
 
+    def test_sequential_flow_interrupted_externally(self):
+        flow = lf.Flow('flow-1').add(
+            utils.ProgressingTask(name='task1'),
+            utils.ProgressingTask(name='task2'),
+            utils.ProgressingTask(name='task3'),
+        )
+        engine = self._make_engine(flow)
+
+        def _run_engine_and_raise():
+            engine_states = {}
+            engine_it = engine.run_iter()
+            while True:
+                try:
+                    engine_state = six.next(engine_it)
+                    if engine_state not in engine_states:
+                        engine_states[engine_state] = 1
+                    else:
+                        engine_states[engine_state] += 1
+                    if engine_states.get(states.SCHEDULING) == 2:
+                        engine_state = engine_it.throw(IOError("I Broke"))
+                        if engine_state not in engine_states:
+                            engine_states[engine_state] = 1
+                        else:
+                            engine_states[engine_state] += 1
+                except StopIteration:
+                    break
+
+        self.assertRaises(IOError, _run_engine_and_raise)
+        self.assertEqual(states.FAILURE, engine.storage.get_flow_state())
+
     def test_sequential_flow_one_task(self):
         flow = lf.Flow('flow-1').add(
             utils.ProgressingTask(name='task1')
@@ -1419,8 +1449,7 @@ class ParallelEngineWithEventletTest(EngineTaskTest,
                      flow_detail=None, executor=None, store=None,
                      **kwargs):
         if executor is None:
-            executor = futurist.GreenThreadPoolExecutor()
-            self.addCleanup(executor.shutdown)
+            executor = 'greenthreads'
         return taskflow.engines.load(flow, flow_detail=flow_detail,
                                      backend=self.backend, engine='parallel',
                                      executor=executor,
@@ -1505,14 +1534,15 @@ class WorkerBasedEngineTest(EngineTaskTest,
         self.worker_thread = tu.daemon_thread(self.worker.run)
         self.worker_thread.start()
 
+        # Ensure worker and thread is stopped when test is done; these are
+        # called in reverse order, so make sure we signal the stop before
+        # performing the join (because the reverse won't work).
+        self.addCleanup(self.worker_thread.join)
+        self.addCleanup(self.worker.stop)
+
         # Make sure the worker is started before we can continue...
         self.worker.wait()
 
-    def tearDown(self):
-        self.worker.stop()
-        self.worker_thread.join()
-        super(WorkerBasedEngineTest, self).tearDown()
-
     def _make_engine(self, flow,
                      flow_detail=None, store=None, **kwargs):
         kwargs.update(self.engine_conf)
